require "ISUI/ISPanelJoypad"
require "ISUI/ISBackButtonWheel"
require "ISUI/ISBuildRecipePanel"
require "JoyPad/JoyPadSetup"
require "ISCustomRadialMenu"

ISBuildingHotbar = ISPanelJoypad:derive("ISBuildingHotbar");
RBButtonHeld = false
DRAG_THRESHOLD_MS = 300  -- time threshold in milliseconds
DRAG_DISTANCE_THRESHOLD = 5  -- pixel threshold

function ISBuildingHotbar:saveData()
    print("Saving BuildingHotbar data")
    local modData = self.character:getModData()
    modData["BuildingHotbarPos"] = { x = self:getX(), y = self:getY() }
    modData["BuildingHotbarRecipes"] = {}
    for i, recipe in ipairs(self.attachedRecipes) do
        if recipe then
            local objInfo = SpriteConfigManager.GetObjectInfo(tostring(recipe:getName()))
            if objInfo then
                modData["BuildingHotbarRecipes"][i] = objInfo:getName()
                print("Saved slot " .. i .. " with object info: " .. objInfo:getName())
            else
                print("Warning: Could not get object info for recipe " .. recipe:getName())
            end
        end
    end
end

function ISBuildingHotbar:loadData()
    local modData = self.character:getModData()
    if modData["BuildingHotbarPos"] then
        self:setX(modData["BuildingHotbarPos"].x)
        self:setY(modData["BuildingHotbarPos"].y)
        print("Loaded BuildingHotbar position: (" .. modData["BuildingHotbarPos"].x .. ", " .. modData["BuildingHotbarPos"].y .. ")")
    end
    if modData["BuildingHotbarRecipes"] then
        for slot, objName in pairs(modData["BuildingHotbarRecipes"]) do
            print("Loading recipe for slot " .. slot .. ": object info " .. objName)
            local objInfo = SpriteConfigManager.GetObjectInfo(objName)
            if objInfo then
                local recipe = objInfo:getRecipe():getCraftRecipe();
                if recipe then
                    self.attachedRecipes[slot] = recipe
                    print("Restored recipe: " .. recipe:getName() .. " to slot " .. slot)
                else
                    print("Warning: No recipe found for object info " .. objInfo:getName())
                end
            else
                print("Warning: Could not retrieve object info for saved name " .. tostring(objName))
            end
        end
    end
end

function ISBuildingHotbar:assignRecipeToSlot(slotIndex, recipe)
    self.attachedRecipes[slotIndex] = recipe;
    self:saveData()
    print("Assigned recipe to slot " .. slotIndex .. ": " .. tostring(recipe:getName()))
    local infos = SpriteConfigManager.GetObjectInfo(tostring(recipe:getName()));
    print(infos)
    self:refresh();
end

function ISBuildingHotbar:getSlotIndexAt(x, y)
    local innerMargin = 15  -- pixels you want to reserve as non-slot area
    -- Only register a slot click if the click is inside the inner area
    if x < innerMargin or x > self.width - innerMargin or y < innerMargin or y > self.height - innerMargin then
        return -1
    end

    local index = math.floor((x - self.margins) / (self.slotWidth + self.slotPad)) + 1
    index = math.max(index, 1)
    return math.min(index, #self.availableSlot)
end

function ISBuildingHotbar:onMouseDown(x, y)
    local slot = self:getSlotIndexAt(x, y)
    -- If click is on a slot icon (close to its center), mark for recipe drag.
    if slot > -1 and self.attachedRecipes[slot] then
        local slotX = self.margins + 1 + (slot - 1) * (self.slotWidth + self.slotPad)
        local slotY = self.margins + 1
        local centerX = slotX + self.slotWidth/2
        local centerY = slotY + self.slotHeight/2
        local tolerance = 40
        local dist = math.sqrt((x - centerX)^2 + (y - centerY)^2)
        if dist <= tolerance then
            self.maybeRecipeDrag = true
            self.dragStartTime = getTimestampMs()
            self.potentialDragSlot = slot
            self.maybeDragRecipe = self.attachedRecipes[slot]
            self.mouseDownX = x
            self.mouseDownY = y
            self.panelDrag = false
            return true
        end
    end
    -- Otherwise, treat as panel drag (click in margin)
    self.panelDrag = true
    self.maybeRecipeDrag = false
    self.mouseDownX = x
    self.mouseDownY = y
    -- Record panel drag offset
    self.dragOffsetX_panel = getMouseX() - self:getX()
    self.dragOffsetY_panel = getMouseY() - self:getY()
    return true
end



function ISBuildingHotbar:onMouseMove(x, y)
    if self.maybeRecipeDrag then
        local currentTime = getTimestampMs()
        local dx = getMouseX() - self.mouseDownX
        local dy = getMouseY() - self.mouseDownY
        if (currentTime - self.dragStartTime >= DRAG_THRESHOLD_MS) or 
           (math.abs(dx) > DRAG_DISTANCE_THRESHOLD or math.abs(dy) > DRAG_DISTANCE_THRESHOLD) then
            -- Start recipe drag: remove the recipe from its slot.
            self.draggingRecipe = self.maybeDragRecipe
            self.draggingSlot = self.potentialDragSlot
            self.attachedRecipes[self.potentialDragSlot] = nil
            self.dragging = true
            -- Compute recipe-drag offset.
            local absX = self:getAbsoluteX()
            local absY = self:getAbsoluteY()
            local slotX = self.margins + 1 + (self.draggingSlot - 1) * (self.slotWidth + self.slotPad)
            local slotY = self.margins + 1
            local iconGlobalX = absX + slotX
            local iconGlobalY = absY + slotY
            local globalMouseX = getMouseX()
            local globalMouseY = getMouseY()
            self.dragOffsetX = globalMouseX - iconGlobalX
            self.dragOffsetY = globalMouseY - iconGlobalY
            self.maybeRecipeDrag = false
        end
        return true
    elseif self.panelDrag then
        -- Update panel dragging using the stored panel offset.
        local currentMouseX = getMouseX()
        local currentMouseY = getMouseY()
        local newX = currentMouseX - self.dragOffsetX_panel
        local newY = currentMouseY - self.dragOffsetY_panel
        self:setX(newX)
        self:setY(newY)
        return true
    elseif self.dragging then
        if self.draggingRecipe then
            self.currentDragX = getMouseX()
            self.currentDragY = getMouseY()
            return true
        end
    end

    return true
end


function ISBuildingHotbar:onMouseMoveOutside(dx, dy)
    if self.panelDrag and self.mouseDownX then
        local newX = self:getX() + dx
        local newY = self:getY() + dy
        self:setX(newX)
        self:setY(newY)
        self:bringToTop()
    elseif self.draggingRecipe then
        -- Update drag coordinates even if mouse is outside.
        self.currentDragX = getMouseX()
        self.currentDragY = getMouseY()
    end
end

function ISBuildingHotbar:onMouseUp(x, y)
    if self.dragging then
        -- Get the panel's absolute boundaries.
        local absX = self:getAbsoluteX()
        local absY = self:getAbsoluteY()
        local globalMouseX = getMouseX()
        local globalMouseY = getMouseY()

        -- Check if the mouse is inside the panel.
        if globalMouseX < absX or globalMouseX > absX + self.width or 
           globalMouseY < absY or globalMouseY > absY + self.height then
            -- Mouse released outside: return dragged recipe to its original slot.
            self.attachedRecipes[self.draggingSlot] = self.draggingRecipe
            print("Mouse released outside; recipe returned to slot " .. self.draggingSlot)
            self:refresh()
        else
            -- Mouse released inside: process normally.
            local targetSlot = self:getSlotIndexAt(x, y)
            if targetSlot > -1 then
                if self.attachedRecipes[targetSlot] then
                    local temp = self.attachedRecipes[targetSlot]
                    self.attachedRecipes[targetSlot] = self.draggingRecipe
                    self.attachedRecipes[self.draggingSlot] = temp
                else
                    self.attachedRecipes[targetSlot] = self.draggingRecipe
                end
                print("Moved recipe to slot " .. targetSlot)
                self:saveData()
                self:refresh()
            else
                self.attachedRecipes[self.draggingSlot] = self.draggingRecipe
                self:refresh()
            end
        end

        -- Clear drag state.
        self.draggingRecipe = nil
        self.draggingSlot = nil
        self.currentDragX = nil
        self.currentDragY = nil
        self.dragging = false
        self.panelDrag = false
        self.mouseDownX = nil
        self.mouseDownY = nil
        self:saveData()
        return true
    elseif self.maybeRecipeDrag then
        local elapsed = getTimestampMs() - self.dragStartTime
        if elapsed < DRAG_THRESHOLD_MS then
            local slot = self.potentialDragSlot
            if slot and self.attachedRecipes[slot] then
                self:buildRecipe(slot, false)
            end
        end
        self.maybeRecipeDrag = false
        self.panelDrag = false
        self.mouseDownX = nil
        self.mouseDownY = nil
        self:saveData()
        return true
    elseif self.panelDrag then
        self.panelDrag = false
        self.mouseDownX = nil
        self.mouseDownY = nil
        self:saveData()
        return true
    else
        local index = self:getSlotIndexAt(x, y)
        if index > -1 then
            self:buildRecipe(index, false)
        end
        self.mouseDownX = nil
        self.mouseDownY = nil
        return true
    end
end

function ISBuildingHotbar:onMouseUpOutside(x, y)
    -- If a recipe was being dragged, return it to its original slot.
    if self.draggingRecipe then
        self.attachedRecipes[self.draggingSlot] = self.draggingRecipe
        print("Mouse released outside; returning recipe to slot " .. self.draggingSlot)
        self:refresh()
    end
    -- Clear all drag state.
    self.draggingRecipe = nil
    self.draggingSlot = nil
    self.currentDragX = nil
    self.currentDragY = nil
    self.dragging = false
    self.maybeDrag = false
    self.panelDrag = false
    self.mouseDownX = nil
    self.mouseDownY = nil
    return true
end

function ISBuildingHotbar:onRightMouseUp(x, y)
    local index = self:getSlotIndexAt(x, y)
    if index > -1 then
         if self.attachedRecipes[index] then
              print("Unassigning recipe from slot " .. index)
              self.attachedRecipes[index] = nil
              self:saveData()
              self:refresh()
         else
              print("No recipe assigned to slot " .. index .. " to unassign")
         end
         return true
    end
    return false
end


function ISBuildingHotbar:render()
    -- Draw the hotbar as before...
    self:drawRectBorderStatic(0, 0, self.width, self.height, 0.8, 0.8, 0.8, 0.8);
    local mouseX = self:getMouseX()
    local mouseY = self:getMouseY()
    local mouseOverSlotIndex = self:getSlotIndexAt(mouseX, mouseY)
    local slotX = self.margins + 1;
    for i, slot in ipairs(self.availableSlot) do
        self:drawRectBorderStatic(slotX, self.margins + 1, self.slotWidth, self.slotHeight, 0.8, 0.8, 0.8, 0.8);
        if i == mouseOverSlotIndex then
            self:drawRect(slotX, self.margins + 1, self.slotWidth, self.slotHeight, 0.2, 1, 1, 1);
        end
        self:drawText(tostring(i), slotX + 3, self.margins + 2, 1, 1, 1, 1, self.font);
        local recipe = self.attachedRecipes[i]
        if recipe then
            local tex = recipe:getIconTexture();
            if tex then
                local texWidth = tex:getWidth();
                local texHeight = tex:getHeight();
                local posX = slotX + (self.slotWidth - texWidth) / 2;
                local posY = self.margins + 1 + (self.slotHeight - texHeight) / 2;
                self:drawTexture(tex, posX, posY, 1, 1, 1, 1);
            else
                print("Warning: Recipe at slot " .. i .. " returned nil texture from getIconTexture()")
            end
        end
        slotX = slotX + self.slotWidth + self.slotPad;
    end

    -- Draw a tooltip if the mouse is hovering over a slot with a recipe.
    if mouseOverSlotIndex > -1 and self.attachedRecipes[mouseOverSlotIndex] and not self.draggingRecipe then
        local recipe = self.attachedRecipes[mouseOverSlotIndex]
        -- For example, get the tooltip string from the recipe.
        local recipeInputs = recipe:getInputs()
        local allInputs = {}
        for i = 0, recipeInputs:size()-1 do
            local input = recipeInputs:get(i)
            local possibleItems = input:getPossibleInputItems()
            local inputAmount = input:getAmount()
            if possibleItems:size() > 0 then
                table.insert(allInputs, { item = possibleItems:get(0), amount = inputAmount })
            end
        end

        -- Optionally, draw a background rectangle behind the tooltip text:
        local titleStr = recipe:getTranslationName()
        local tooltipLines = {}
        -- table.insert(tooltipLines, titleStr)  -- Assuming titleStr is your recipe title
        table.insert(tooltipLines, " Items Required:")  -- Assuming titleStr is your recipe title
        for i = 1, #allInputs do
            local entry = allInputs[i]
            local line = "  " .. entry.amount .. "x " .. entry.item:getDisplayName()
            table.insert(tooltipLines, line)
        end

        local fullTooltip = table.concat(tooltipLines, "\n")
        local FONT_HGT_SMALL = getTextManager():getFontHeight(UIFont.Small)
        local FONT_HGT_MEDIUM = getTextManager():getFontHeight(UIFont.Medium)
        local globalMouseX = getMouseX()
        local globalMouseY = getMouseY()

        -- Calculate desired tooltip position in global coordinates
        local tooltipX = globalMouseX + 30
        local tooltipY = globalMouseY - (FONT_HGT_SMALL + 60)

        local screenWidth = getCore():getScreenWidth()
        local screenHeight = getCore():getScreenHeight()

        local tooltipWidth = 200  -- adjust as needed
        local tooltipHeight = FONT_HGT_MEDIUM + FONT_HGT_SMALL * #tooltipLines + 10

        -- Clamp the global tooltip position to the screen boundaries
        if tooltipX + tooltipWidth > screenWidth then
            tooltipX = screenWidth - tooltipWidth
        end
        if tooltipY + tooltipHeight > screenHeight then
            tooltipY = screenHeight - tooltipHeight
        end
        if tooltipX < 0 then tooltipX = 0 end
        if tooltipY < 0 then tooltipY = 0 end

        -- Convert the global coordinates to local coordinates for this panel
        local panelAbsX = self:getAbsoluteX()
        local panelAbsY = self:getAbsoluteY()
        local localTooltipX = tooltipX - panelAbsX
        local localTooltipY = tooltipY - panelAbsY

        -- Now draw the tooltip
        self:drawRect(localTooltipX - 2, localTooltipY - 2, tooltipWidth, tooltipHeight, 0.7, 0, 0, 0)
        self:drawText(titleStr, localTooltipX, localTooltipY, 1, 1, 1, 1, UIFont.Medium)
        self:drawText(fullTooltip, localTooltipX, localTooltipY + 26, 1, 1, 1, 1, UIFont.Small)
    end

    -- Draw dragged recipe icon (if any)
    if self.draggingRecipe and self.currentDragX and self.currentDragY then
        local tex = self.draggingRecipe:getIconTexture()
        if tex then
            local globalIconX = self.currentDragX - self.dragOffsetX
            local globalIconY = self.currentDragY - self.dragOffsetY
            local absX = self:getAbsoluteX()
            local absY = self:getAbsoluteY()
            local localX = globalIconX - absX
            local localY = globalIconY - absY
            self:drawTexture(tex, localX, localY, 1, 1, 1, 1)
        end
    end
end




function ISBuildingHotbar:new(character)
    local numSlots = 8
    local slotWidth = 60
    local slotPad = 10
    local margins = 15
    local totalWidth = (numSlots * slotWidth) + ((numSlots - 1) * slotPad) + (2 * margins)

    local o = ISPanelJoypad:new(100, 100, totalWidth, 90);
    setmetatable(o, self);
    self.__index = self;
    o.logic = BuildLogic.new(character, nil, nil);
    o.character = character;
    o.slotWidth = slotWidth;
    o.slotHeight = 60;
    o.slotPad = slotPad;
    o.margins = margins;
    o.anchorLeft = false;
    o.anchorRight = false;
    o.anchorTop = false;
    o.anchorBottom = false;
    o.isJoypad = false;
    o.availableSlot = {};
    o.attachedRecipes = {};
    o.selectedRecipe = nil;
    o.panelDrag = false
    o.draggingRecipe = nil  -- the recipe being dragged
    o.draggingSlot = nil    -- original slot index
    o.currentDragX = nil    -- current drag X position
    o.currentDragY = nil    -- current drag Y position
    o.maybeDrag = false
    o.dragStartTime = 0
    o.potentialDragSlot = nil
    o.maybeDragRecipe = nil

    for i = 1, numSlots do
        o.availableSlot[i] = { name = "Recipe Slot " .. i };
        o.attachedRecipes[i] = nil;
    end
    o.font = UIFont.Small;
    o.assignmentPrompt = {};
    o:loadData()
    return o;
end

-- Function to attach a building recipe to a slot.
function ISBuildingHotbar:attachRecipe(recipe, slotIndex)
    if slotIndex < 1 or slotIndex > #self.availableSlot then
        return;
    end
    self.attachedRecipes[slotIndex] = recipe;
end

function ISBuildingHotbar:showAssignmentPrompt()
    local screenW = getCore():getScreenWidth()
    local screenH = getCore():getScreenHeight()
    local width = 390
    local height = 80
    local x = (screenW - width) / 2
    local y = (screenH - height) / 2

    local prompt = ISRichTextPanel:new(x, y, width, height)
    prompt:setText("Press 1-8 to assign recipe to Builder Hotbar")
    prompt:initialise()
    prompt:instantiate()
    prompt:setVisible(true)
    prompt:setMargins(16, 16, 16, 16)
    prompt:setContentTransparency(1.0)
    prompt.borderColor = { r = 1, g = 1, b = 1, a = 1 }
    prompt.backgroundColor = { r = 0, g = 0, b = 0, a = 0.7 }
    prompt.defaultFont = UIFont.NewLarge
    prompt.autosetheight = false
    prompt.font = prompt.defaultFont
    prompt:paginate()
    prompt:addToUIManager()
    local buttonWidth = 60
    local buttonHeight = 20
    local buttonX = width / 2 - 30
    local buttonY = prompt:getHeight() - buttonHeight - 5
    local cancelButton = ISButton:new(buttonX, buttonY, buttonWidth, buttonHeight, "Cancel", self, ISBuildingHotbar.onCancelAssignment)
    cancelButton:initialise()
    cancelButton:instantiate()
    prompt:addChild(cancelButton)
    self.assignmentPrompt = prompt
end

function ISBuildingHotbar.onCancelAssignment(self, button)
    if ISCustomBuildWindow then
        BuildingAssignmentMode = false
        BuildingAssignmentRecipe = nil
        if self.assignmentPrompt then
            self.assignmentPrompt:removeFromUIManager()
            self.assignmentPrompt = nil
        end
    end
end

function ISBuildingHotbar:activateSlot(slotIndex)
    if BuildingAssignmentMode then
        self:assignRecipeToSlot(slotIndex, BuildingAssignmentRecipe)
        print("Assigned pending recipe to slot " .. slotIndex)
        BuildingAssignmentMode = false
        BuildingAssignmentRecipe = nil
        self.assignmentPrompt:removeFromUIManager()
        self.assignmentPrompt = nil
        return
    end

    local recipe = self.attachedRecipes[slotIndex]
    if not recipe then return; end
    print("Pressed index " .. slotIndex)
    if recipe.activate and type(recipe.activate) == "function" then
        recipe:activate(self.character)
    else
        print("Warning: Recipe " .. recipe:getName() .. " has no activate method.")
    end
end


function ISBuildingHotbar:buildRecipe(slotIndex, repeated, faceDir)
    local player = self.character
    if not self.isJoypad and not self:getIsVisible() then return end
    local recipe = self.attachedRecipes[slotIndex]
    local containers = ISInventoryPaneContextMenu.getContainers(player)
    self.logic:setRecipe(recipe)
    self.logic:setContainers(containers)
    local canBuild = self.logic:canPerformCurrentRecipe()
    local cheat = self.character:isBuildCheat()
    if not recipe then
        print("No recipe assigned to slot " .. slotIndex)
        return
    end
    if not cheat and not canBuild then
        if repeated then
            return
        end
        player:Say("I don't have all the materials for that!")
        return
    end
    print("Building recipe from slot " .. slotIndex .. ": " .. recipe:getName())

    local resultItem = recipe:getName()
    if not resultItem then
        print("Recipe " .. recipe:getName() .. " does not provide a result identifier!")
        return
    end

    local objectInfo = SpriteConfigManager.GetObjectInfo(resultItem)
    if not objectInfo then
        print("Could not retrieve objectInfo for result: " .. tostring(resultItem))
        return
    end
    if faceDir == "w" then
        faceDir = 1
    elseif faceDir == "n" then
        faceDir = 2
    elseif faceDir == "e" then
        faceDir = 3
    elseif faceDir == "s" then
        faceDir = 4
    elseif faceDir == "single" then
        faceDir = 1
    elseif not faceDir then
        faceDir = 1
    end

    local buildEntity = ISCustomBuildIsoEntity:new(player, objectInfo, faceDir, containers, slotIndex)
    if not buildEntity then
         print("Failed to create build entity")
         return
    end

    getCell():setDrag(buildEntity, player:getPlayerNum())

    local square = player:getCurrentSquare()
    if square then
         DoTileBuilding(buildEntity, true, square:getX(), square:getY(), square:getZ(), square)
    end
    print("After DoTileBuilding")
end


function ISBuildingHotbar:refresh()
    local numSlots = #self.availableSlot;
    local totalWidth = numSlots * self.slotWidth + (numSlots - 1) * self.slotPad + self.margins * 2;
    self:setWidth(totalWidth);

    self:setHeight(self.slotHeight + self.margins * 2);

    for i, slot in ipairs(self.availableSlot) do
        local recipe = self.attachedRecipes[i];
        if recipe and recipe.getIconTexture then
            slot.texture = recipe:getIconTexture();
        else
            slot.texture = nil;
        end
    end
end

function ISBuildingHotbar:getSlotForKey(key)
    if getCore():isKey("Hotbar 1", key) then return 1; end
    if getCore():isKey("Hotbar 2", key) then return 2; end
    if getCore():isKey("Hotbar 3", key) then return 3; end
    if getCore():isKey("Hotbar 4", key) then return 4; end
    if getCore():isKey("Hotbar 5", key) then return 5; end
    if getCore():isKey("Hotbar 6", key) then return 6; end
    if getCore():isKey("Hotbar 7", key) then return 7; end
    if getCore():isKey("Hotbar 8", key) then return 8; end
    return -1;
end

function ISBuildingHotbar:customUIKeyHandler(key)
    local slotIndex = self:getSlotForKey(key)
    if slotIndex ~= -1 then
        if Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) then
            self:buildRecipe(slotIndex, false)
            return true
        end
    end
    if Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) then
        if getCore():isKey("Building Hotbar", key) then
            if ISBuildingHotbar.instance and ISBuildingHotbar.instance:getIsVisible() then
                ISBuildingHotbar.instance:removeFromUIManager()
                ISBuildingHotbar.instance:setVisible(false)
                print("Building Hotbar closed.")
            else
                local player = getSpecificPlayer(0)
                if not ISBuildingHotbar.instance then
                    ISBuildingHotbar.instance = ISBuildingHotbar:new(player)
                end
                ISBuildingHotbar.instance:addToUIManager()
                ISBuildingHotbar.instance:setVisible(true)
                print("Building Hotbar opened.")
            end
            return true
        end
    end
    return false
end

ISBuildingHotbar.onKeyStartPressed = function(key)
    local playerObj = getSpecificPlayer(0)
    if not ISBuildingHotbar.instance or not playerObj or playerObj:isDead() then
        return
    end
    if UIManager.getSpeedControls() and UIManager.getSpeedControls():getCurrentGameSpeed() == 0 then
        return
    end
    if JoypadState.players[1] then
        return
    end
    local self = ISBuildingHotbar.instance
    local slotToCheck = self:getSlotForKey(key)
    if slotToCheck == -1 then return end
    local radialMenu = getPlayerRadialMenu(0)
    if getCore():getOptionRadialMenuKeyToggle() and radialMenu:isReallyVisible() then
        self.radialWasVisible = true
        radialMenu:removeFromUIManager()
        return
    end
    self.keyPressedMS = getTimestampMs()
    self.radialWasVisible = false
end

local originalHotbarKeyPressed = ISHotbar.onKeyPressed

ISHotbar.onKeyPressed = function(key)
    local buildingHotbar = ISBuildingHotbar.instance
    if buildingHotbar and buildingHotbar:getIsVisible() and Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) then
        local playerObj = getSpecificPlayer(0)
        if not playerObj or playerObj:isDead() then return end
        if UIManager.getSpeedControls() and UIManager.getSpeedControls():getCurrentGameSpeed() == 0 then return end
        if JoypadState.players[1] then return end

        local validKey = false
        if getCore():isKey("Hotbar 1", key) then validKey = true end
        if getCore():isKey("Hotbar 2", key) then validKey = true end
        if getCore():isKey("Hotbar 3", key) then validKey = true end
        if getCore():isKey("Hotbar 4", key) then validKey = true end
        if getCore():isKey("Hotbar 5", key) then validKey = true end
        if getCore():isKey("Hotbar 6", key) then validKey = true end
        if getCore():isKey("Hotbar 7", key) then validKey = true end
        if getCore():isKey("Hotbar 8", key) then validKey = true end
        if not validKey then return end

        local slotToCheck = buildingHotbar:getSlotForKey(key)
        if slotToCheck == -1 then return end

        local radialMenu = getPlayerRadialMenu(0)
        if radialMenu:isReallyVisible() or buildingHotbar.radialWasVisible then
            if not getCore():getOptionRadialMenuKeyToggle() then
                radialMenu:removeFromUIManager()
            end
            return
        end
        if playerObj:isAttacking() then return end
        local queue = ISTimedActionQueue.queues[playerObj]
        if queue and #queue.queue > 0 then return end

        if Keyboard.isKeyDown(Keyboard.KEY_LSHIFT) then
            buildingHotbar:buildRecipe(slotToCheck)
        else
            buildingHotbar:activateSlot(slotToCheck)
        end
    else
        originalHotbarKeyPressed(key)
    end
end


local originalCallCommand = ISBackButtonWheel.onCommand

function ISBackButtonWheel:onCommand(command)
    if command == "Building" then
        ISBuildingHotbar.onBackButtonWheelActivated()
    else
        originalCallCommand(self, command)
    end
end

local originalBuildRecipe = ISBuildRecipePanel.createDynamicChildren

function ISBuildRecipePanel:createDynamicChildren()
    local recipe = self.logic:getRecipe();
    ISBuildingHotbar.instance.selectedRecipe = recipe
    originalBuildRecipe(self)
end

local originalHoldButton = JoypadControllerData.onHoldButton

function JoypadControllerData:onHoldButton(button, time)
    if not ISCustomBuildWindow.instance then
        if button == 5 and not RBButtonHeld then
            RBButtonHeld = true
        elseif button == 10 and RBButtonHeld and not ISCustomRadialMenu.instance then
            local radialBuildMenu = ISCustomRadialMenu:new(getSpecificPlayer(0), "build")
            radialBuildMenu:fillMenu()
            return
        else
            originalHoldButton(self, button, time)
        end
    elseif ISCustomBuildWindow.instance then
        if button == 2 and not ISCustomRadialMenu.instance then
            local playerObj = getSpecificPlayer(0)
            local radialBuildMenu = ISCustomRadialMenu:new(playerObj, "assign")
	        radialBuildMenu:fillMenu()
            return
        else
            originalHoldButton(self, button, time)
            return
        end
    end
end

local originalReleaseButton = JoypadControllerData.onReleaseButton

function JoypadControllerData:onReleaseButton(button)
    if not ISCustomBuildWindow.instance and ISCustomRadialMenu.instance then
        if tostring(button) == "10" then
            RBButtonHeld = false
            return
        elseif tostring(button) == "5" then
            RBButtonHeld = false
            return
        else
            RBButtonHeld = false
            originalReleaseButton(self, button)
        end
    else
        RBButtonHeld = false
        originalReleaseButton(self, button)
    end
end

local function createWindow(_player, _windowInstance, _isoObject)
    local playerNum = _player:getPlayerNum();

    local x = getMouseX() + 10;
    local y = getMouseY() + 10;
    local adjustPos = true;

    local width = 0;
    local height = 0;

    local windowKey = _windowInstance.xuiStyleName or "Default";

    -- close all other entity windows - we only allow one open at the moment. - spurcival
    ISEntityUI.CloseWindows();

    if ISEntityUI.players[playerNum] and ISEntityUI.players[playerNum].windows[windowKey] then
        if ISEntityUI.players[playerNum].windows[windowKey].instance then
            ISEntityUI.players[playerNum].windows[windowKey].instance:close();
        end
        if ISEntityUI.players[playerNum].windows[windowKey].x and ISEntityUI.players[playerNum].windows[windowKey].y then
            x = ISEntityUI.players[playerNum].windows[windowKey].x;
            y = ISEntityUI.players[playerNum].windows[windowKey].y;
            adjustPos = false;
        end
        if ISEntityUI.players[playerNum].windows[windowKey].width and ISEntityUI.players[playerNum].windows[windowKey].height then
            width = ISEntityUI.players[playerNum].windows[windowKey].width;
            height = ISEntityUI.players[playerNum].windows[windowKey].height;
        end
    else
        ISEntityUI.players[playerNum] = ISEntityUI.players[playerNum] or {};
        ISEntityUI.players[playerNum].windows = ISEntityUI.players[playerNum].windows or {};
        ISEntityUI.players[playerNum].windows[windowKey] = {};
    end

    _windowInstance:initialise();
    _windowInstance:instantiate();
    _windowInstance:setX(x);
    _windowInstance:setY(y);
    _windowInstance:setVisible(true);
    if _windowInstance.calculateLayout then
        _windowInstance:calculateLayout(width,height);
    end
    _windowInstance:addToUIManager();

    ISEntityUI.players[playerNum].windows[windowKey].instance = _windowInstance;
    ISEntityUI.players[playerNum].windows[windowKey].playerObj = _player;
    ISEntityUI.players[playerNum].windows[windowKey].entityObj = _isoObject;

    --first time open panel and isoobject then middle of screen.
    if adjustPos and instanceof(_isoObject, "IsoObject") then
        local x = (getCore():getScreenWidth()/2) - (_windowInstance:getWidth()/2);
        local y = (getCore():getScreenHeight()/2) - (_windowInstance:getHeight()/2);
        _windowInstance:setX(x);
        _windowInstance:setY(y);
        ISEntityUI.players[playerNum].windows[windowKey].x = x;
        ISEntityUI.players[playerNum].windows[windowKey].y = y;
    end

    if JoypadState.players[playerNum+1] then
        if getFocusForPlayer(playerNum) then getFocusForPlayer(playerNum):setVisible(false); end
        if getPlayerInventory(playerNum) then getPlayerInventory(playerNum):close(); end
        if getPlayerLoot(playerNum) then getPlayerLoot(playerNum):close(); end
        --setJoypadFocus(playerNum, nil);
        setJoypadFocus(playerNum, _windowInstance);
    end
end

ISBuildingHotbar.onBackButtonWheelActivated = function()
    local _player = getSpecificPlayer(0);
    if not ISCustomBuildWindow.instance then
        local windowInstance = ISXuiSkin.build(skin, "BuildWindow", ISCustomBuildWindow, 0, 0, 60, 30, _player, nil, '*');
        createWindow(_player, windowInstance, nil);
        ISCustomBuildWindow.instance = windowInstance;
    else
        ISCustomBuildWindow.instance:close();
        ISCustomRadialMenu.instance = nil
    end
end


Events.OnGameStart.Add(function()
    local player = getSpecificPlayer(0);
    Events.OnKeyStartPressed.Add(ISBuildingHotbar.onKeyStartPressed)
    Events.OnKeyPressed.Add(ISBuildingHotbar.onKeyPressed)
    ISBuildingHotbar.instance = ISBuildingHotbar:new(player);
    ISBuildingHotbar.instance:addToUIManager();
    if JoypadState.players[1] then
        ISBuildingHotbar.instance:setVisible(false);
        ISBuildingHotbar.instance.isJoypad = true;
    end

    Events.OnCustomUIKey.Add(function(key)
         if ISBuildingHotbar.instance then
              if ISBuildingHotbar.instance:customUIKeyHandler(key) then
                   return true;
              end
         end
         return false;
    end)
end)